@@ -19,6 +19,10 @@
 #ifndef MICRODNS_COMPAT_H
 #define MICRODNS_COMPAT_H
 
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
 enum {
         MDNS_STDERR = -1, // standard error
         MDNS_NETERR = -2, // network error
@@ -95,15 +99,10 @@ extern const char *inet_ntop(int af, const void *src, char *dst, socklen_t size)
 extern int inet_pton(int af, const char *src, void *dst);
 # endif // !inet_ntop
 
-typedef DWORD multicast_if;
-
 #else
 
 # if HAVE_IFADDRS_H
 #include <ifaddrs.h>
-typedef struct sockaddr_storage multicast_if;
-# else
-typedef void* multicast_if;
 # endif
 
 #endif // _WIN32
@@ -133,7 +132,18 @@ struct pollfd
 int poll(struct pollfd *fds, unsigned nfds, int timeout);
 #endif
 
+#if defined(_MSC_VER)
+#include <io.h>
+#include <basetsd.h>
+
+typedef SSIZE_T ssize_t;
+
+#define alloca(s) _alloca(s)
+#define write(f,b,c) _write(f,b,c)
+#endif
+
 extern int os_strerror(int, char *, size_t);
-extern int os_mcast_join(sock_t, const struct sockaddr_storage *, multicast_if mintf);
+extern int os_mcast_join(sock_t, const struct sockaddr_storage *, uint32_t intf_idx,
+                         const struct sockaddr_storage* intf_addr);
 
 #endif /* MICRODNS_COMPAT_H */
