@@ -16,11 +16,17 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include <compat.h>
 #include <stdio.h>
-#include <unistd.h>
 #include <signal.h>
 
-#include "microdns.h"
+#include <microdns/microdns.h>
+
+#include "compat.h"
+
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
 
 volatile sig_atomic_t sigflag = 0;
 
@@ -71,7 +77,7 @@ int main(int i_argc, char *ppsz_argv[])
         }
         signal(SIGINT, &sighandler);
 
-        if ((r = mdns_init(&ctx, MDNS_ADDR_IPV4, MDNS_PORT)) < 0)
+        if ((r = mdns_init(&ctx, NULL, MDNS_PORT)) < 0)
                 goto err;
         if ((r = mdns_listen(ctx, ppsz_names, i_nb_names, RR_PTR, 10, stop,
                              callback, NULL)) < 0)
