@@ -36,12 +36,12 @@
 
 #include "compat.h"
 #include "utils.h"
-#include "microdns.h"
+#include "microdns/microdns.h"
+#include "mdns.h"
 
 #define MDNS_PKT_MAXSZ 4096 // read/write buffer size
 
 struct mdns_svc {
-        char *name;
         enum rr_type type;
         union
         {
@@ -54,23 +54,23 @@ struct mdns_svc {
 
 struct mdns_conn {
         sock_t sock;
-        multicast_if if_addr;  // NB: In Windows this is the interface index
-        struct mdns_ip mdns_ip;  // IP address and family of the interface
+        uint32_t intf_idx;
+        struct sockaddr_storage intf_addr;  // IP address and family of the interface
+        struct sockaddr_storage mcast_addr; // The multicast address targeted by this connection
 };
 
 struct mdns_ctx {
         struct mdns_conn *conns;
         size_t nb_conns;
-        struct sockaddr_storage addr;
         struct mdns_svc *services;
 };
 
 static int mdns_resolve(struct mdns_ctx *ctx, const char *addr, unsigned short port);
-static size_t mdns_write_hdr(uint8_t *, const struct mdns_hdr *);
+static size_t mdns_write_hdr(uint8_t *, size_t *s, const struct mdns_hdr *);
 static int strrcmp(const char *, const char *);
 
 extern const uint8_t *rr_read(const uint8_t *, size_t *, const uint8_t *, struct rr_entry *, int8_t ans);
-extern size_t rr_write(uint8_t *, const struct rr_entry *, int8_t ans);
+extern ssize_t rr_write(uint8_t *, size_t *, const struct rr_entry *, int8_t ans);
 extern void rr_print(const struct rr_entry *);
 extern void rr_free(struct rr_entry *);
 
@@ -78,38 +78,59 @@ extern void rr_free(struct rr_entry *);
 #if HAVE_GETIFADDRS
 
 static bool
-mdns_is_interface_valuable(struct ifaddrs* ifa)
+mdns_is_interface_valuable(const struct ifaddrs* ifa, int family)
 {
-    return ifa->ifa_addr != NULL && (ifa->ifa_addr->sa_family == AF_INET ||
-                                     ifa->ifa_addr->sa_family == AF_INET6) &&
+    struct sockaddr_in6 saddr;
+
+    if (ifa->ifa_addr == NULL)
+      return false;
+
+    memcpy(&saddr, ifa->ifa_addr, sizeof(saddr));
+
+    return ifa->ifa_addr->sa_family == family &&
             (ifa->ifa_flags & IFF_LOOPBACK) == 0 &&
             (ifa->ifa_flags & IFF_UP) != 0 &&
             (ifa->ifa_flags & IFF_RUNNING) != 0 &&
-            ((ifa->ifa_addr->sa_family == AF_INET6 &&
-                    ((struct sockaddr_in6*)(ifa->ifa_addr))->sin6_scope_id == 0) ||
+            /* We only want a link local address, especially since we then
+             * can get the associated interface index */
+            ((family == AF_INET6 && saddr.sin6_scope_id != 0) ||
                 ifa->ifa_addr->sa_family == AF_INET);
 }
 
+static size_t count_interfaces(const struct ifaddrs *ifs,
+                               const struct addrinfo* addrs)
+{
+    size_t nb_if = 0;
+
+    for ( const struct addrinfo* addr = addrs; addr != NULL; addr = addr->ai_next ) {
+            for (const struct ifaddrs *c = ifs; c != NULL; c = c->ifa_next) {
+                    if (!mdns_is_interface_valuable(c, addr->ai_family)) {
+                            continue;
+                    }
+                    nb_if++;
+            }
+    }
+    return nb_if;
+}
+
 static int
-mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size_t* p_nb_intf, int ai_family)
+mdns_list_interfaces(uint32_t** pp_intfs, struct sockaddr_storage **pp_mdns_ips,
+                     size_t* p_nb_intf, struct sockaddr_storage **pp_mcast_addrs,
+                     const struct addrinfo* addrs)
 {
         struct ifaddrs *ifs;
-        struct mdns_ip *mdns_ips;
+        struct sockaddr_storage *mdns_ips;
         struct ifaddrs *c;
+        struct sockaddr_storage *mcast_addrs;
         size_t nb_if;
-        multicast_if* intfs;
+        uint32_t* intfs;
 
         *p_nb_intf = 0;
         if (getifaddrs(&ifs) || ifs == NULL)
                 return (MDNS_NETERR);
-        nb_if = 0;
-        for (c = ifs; c != NULL; c = c->ifa_next) {
-                if (c->ifa_addr == NULL ||
-                    c->ifa_addr->sa_family != ai_family ||
-                    !mdns_is_interface_valuable(c))
-                        continue;
-                nb_if++;
-        }
+
+        nb_if = count_interfaces(ifs, addrs);
+
         if (nb_if == 0) {
                 freeifaddrs(ifs);
                 return (MDNS_ERROR);
@@ -122,25 +143,35 @@ mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size
         *pp_mdns_ips = mdns_ips = malloc(sizeof(*mdns_ips) * nb_if);
         if (mdns_ips == NULL) {
                 free(intfs);
+                *pp_intfs = NULL;
                 freeifaddrs(ifs);
                 return (MDNS_ERROR);
         }
-        for (c = ifs; c != NULL; c = c->ifa_next) {
-                if (c->ifa_addr == NULL ||
-                    c->ifa_addr->sa_family != ai_family ||
-                    !mdns_is_interface_valuable(c))
-                        continue;
-                memcpy(intfs, c->ifa_addr, sizeof(*intfs));
-                if (ai_family == AF_INET) {
-                        struct sockaddr_in *addr_in = (struct sockaddr_in *)c->ifa_addr;
-                        memcpy(&(*mdns_ips).ipv4.addr, &addr_in->sin_addr, sizeof(struct in_addr));
+        *pp_mcast_addrs = mcast_addrs = malloc(sizeof(*mcast_addrs) * nb_if);
+        if (mcast_addrs == NULL) {
+                free(mdns_ips);
+                free(intfs);
+                *pp_intfs = NULL;
+                *pp_mdns_ips = NULL;
+                freeifaddrs(ifs);
+                return (MDNS_ERROR);
+        }
+        for ( const struct addrinfo* addr = addrs; addr != NULL; addr = addr->ai_next ) {
+                for (c = ifs; c != NULL; c = c->ifa_next) {
+                        if (!mdns_is_interface_valuable(c, addr->ai_family))
+                                continue;
+                        if (c->ifa_addr->sa_family == AF_INET) {
+                                *intfs = if_nametoindex(c->ifa_name);
+                        } else {
+                                memcpy(intfs, &((struct sockaddr_in6*)c->ifa_addr)->sin6_scope_id,
+                                       sizeof(*intfs));
+                        }
+                        memcpy(mdns_ips, c->ifa_addr, sa_len(c->ifa_addr));
+                        memcpy(mcast_addrs, addr->ai_addr, sa_len(addr->ai_addr));
+                        mdns_ips++;
+                        intfs++;
+                        mcast_addrs++;
                 }
-                else {
-                        struct sockaddr_in6 *addr_in = (struct sockaddr_in6 *)c->ifa_addr;
-                        memcpy(&(*mdns_ips).ipv6.addr, &addr_in->sin6_addr, sizeof(struct in6_addr));
-                }
-                mdns_ips++;
-                intfs++;
         }
         freeifaddrs(ifs);
         *p_nb_intf = nb_if;
@@ -148,21 +179,56 @@ mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size
 }
 #else
 static size_t
-mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size_t* p_nb_intf, int ai_family)
+mdns_list_interfaces(uint32_t** pp_intfs, struct sockaddr_storage **pp_mdns_ips,
+                     size_t* p_nb_intf, struct sockaddr_storage **pp_mcast_addrs,
+                     const struct addrinfo* addrs)
 {
-        multicast_if *intfs;
-        struct mdns_ip *mdns_ips;
-        *pp_intfs = intfs = malloc(sizeof(*intfs));
+        struct sockaddr_storage *mdns_ips;
+        uint32_t *intfs;
+        size_t nb_intfs = 0;
+        for( const struct addrinfo* addr = addrs; addr != NULL; addr = addr->ai_next ) {
+                if( addr->ai_family == AF_INET || addr->ai_family == AF_INET6 )
+                    ++nb_intfs;
+        }
+        *pp_intfs = intfs = malloc( nb_intfs * sizeof(*intfs) );
         if (intfs == NULL)
                 return (MDNS_ERROR);
-        memset(intfs, 0, sizeof(*intfs));
-        *pp_mdns_ips = mdns_ips = malloc(sizeof(*mdns_ips));
+        *pp_mdns_ips = mdns_ips = malloc( nb_intfs * sizeof(*mdns_ips) );
         if (mdns_ips == NULL) {
-                free(mdns_ips);
+                free(intfs);
+                *pp_intfs = NULL;
                 return (MDNS_ERROR);
         }
-        memset(mdns_ips, 0, sizeof(*mdns_ips));
-        *p_nb_intf = 1;
+
+        struct sockaddr_storage* mcast_addrs;
+        mcast_addrs = *pp_mcast_addrs = malloc( nb_intfs * sizeof( *mcast_addrs ) );
+        if( mcast_addrs == NULL ) {
+            free(mdns_ips);
+            free(intfs);
+            *pp_intfs = NULL;
+            *pp_mdns_ips = NULL;
+            return (MDNS_ERROR);
+        }
+        for( const struct addrinfo* addr = addrs; addr != NULL; addr = addr->ai_next ) {
+                if( addr->ai_family == AF_INET || addr->ai_family == AF_INET6 ) {
+                        *intfs = 0;
+                        /* Take a shortcut assuming when we are in this case, we
+                         * can use MCAST_JOIN_GROUP. Otherwise, we need the
+                         * adapter IP address, which without getifaddrs is a pain
+                         * in the lower back to fetch
+                         */
+#ifndef MCAST_JOIN_GROUP
+# error We need the adapter address
+#endif
+                        mdns_ips->ss_family = addr->ai_family;
+                        memcpy(mcast_addrs, addr->ai_addr, sa_len(addr->ai_addr));
+                        intfs++;
+                        mdns_ips++;
+                        mcast_addrs++;
+                }
+        }
+
+        *p_nb_intf = nb_intfs;
         return (0);
 }
 #endif // HAVE_GETIFADDRS
@@ -170,17 +236,22 @@ mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size
 #else // _WIN32
 
 static bool
-mdns_is_interface_valuable(IP_ADAPTER_ADDRESSES *intf)
+mdns_is_interface_valuable(IP_ADAPTER_ADDRESSES *intf, int family)
 {
     return (intf->IfType == IF_TYPE_IEEE80211 || intf->IfType == IF_TYPE_ETHERNET_CSMACD) &&
-            intf->OperStatus == IfOperStatusUp;
+            intf->OperStatus == IfOperStatusUp &&
+            ((family == AF_INET && intf->Ipv4Enabled) ||
+            (family == AF_INET6 && intf->Ipv6Enabled));
 }
 
 static size_t
-mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size_t* p_nb_intf, int ai_family)
+mdns_list_interfaces(uint32_t** pp_intfs, struct sockaddr_storage **pp_mdns_ips,
+                     size_t* p_nb_intf, struct sockaddr_storage **pp_mcast_addrs,
+                     const struct addrinfo* addrs)
 {
-        multicast_if* intfs;
-        struct mdns_ip *mdns_ips;
+        uint32_t* intfs;
+        struct sockaddr_storage *mdns_ips;
+        struct sockaddr_storage *mcast_addrs;
         IP_ADAPTER_ADDRESSES *res = NULL, *current;
         ULONG size;
         HRESULT hr;
@@ -203,7 +274,7 @@ mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size
                 res = malloc( size );
                 if (res == NULL)
                         return (MDNS_ERROR);
-                hr = GetAdaptersAddresses(ai_family, GAA_FLAG_SKIP_ANYCAST |
+                hr = GetAdaptersAddresses(AF_UNSPEC, GAA_FLAG_SKIP_ANYCAST |
                                                     GAA_FLAG_SKIP_DNS_SERVER,
                                                     NULL, res, &size);
         } while (hr == ERROR_BUFFER_OVERFLOW);
@@ -212,10 +283,12 @@ mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size
                 return (MDNS_NETERR);
         }
 
-        for (current = res; current != NULL; current = current->Next) {
-                if (!mdns_is_interface_valuable(current))
-                        continue;
-                ++nb_intf;
+        for ( const struct addrinfo* addr = addrs; addr != NULL; addr = addr->ai_next ) {
+                for (current = res; current != NULL; current = current->Next) {
+                        if (!mdns_is_interface_valuable(current, addr->ai_family))
+                                continue;
+                        ++nb_intf;
+                }
         }
         if (nb_intf == 0) {
                 // Fallback to the default interface
@@ -225,13 +298,21 @@ mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size
                         return (MDNS_ERROR);
                 }
                 **pp_intfs = 0;
-                *pp_mdns_ips = malloc(sizeof(*mdns_ips));
+                *pp_mdns_ips = mdns_ips = malloc(sizeof(*mdns_ips));
                 if (*pp_mdns_ips == NULL) {
                         free(*pp_intfs);
                         free(res);
                         return (MDNS_ERROR);
                 }
                 memset(mdns_ips, 0, sizeof(*mdns_ips));
+                *pp_mcast_addrs = mcast_addrs = malloc(sizeof(*mcast_addrs));
+                if (mcast_addrs == NULL) {
+                        free(mdns_ips);
+                        free(*pp_intfs);
+                        free(res);
+                        return (MDNS_ERROR);
+                }
+                memcpy(mcast_addrs, addrs->ai_addr, sa_len(addrs->ai_addr));
                 *p_nb_intf = 1;
                 return (0);
         }
@@ -247,36 +328,50 @@ mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size
                 free(res);
                 return (MDNS_ERROR);
         }
-        for (current = res; current != NULL; current = current->Next) {
-                if (!mdns_is_interface_valuable(current))
-                        continue;
-                if (ai_family == AF_INET6) {
-                        // For IPv6, The input value for setting IPV6_MULTICAST_IF is the
-                        // interface index of the desired outgoing interface in *host byte order*.
-                        // See https://docs.microsoft.com/en-us/windows/win32/winsock/ipproto-ipv6-socket-options
-                        *intfs = current->Ipv6IfIndex;
-                }
-                else {
-                        // For IPv4, The input value for setting IP_MULTICAST_IF is the
-                        // interface index of the desired outgoing interface in *network byte order*.
-                        // See https://docs.microsoft.com/en-us/windows/win32/winsock/ipproto-ip-socket-options
-                        *intfs = htonl(current->IfIndex);
-                }
-                PIP_ADAPTER_UNICAST_ADDRESS_XP p_unicast = current->FirstUnicastAddress;
-                if( p_unicast )
-                {
-                        // Take the first unicast address (highest priority)
-                        if (ai_family == AF_INET) {
-                                struct sockaddr_in *addr_in = (struct sockaddr_in*) p_unicast->Address.lpSockaddr;
-                                memcpy(&(*mdns_ips).ipv4.addr, &addr_in->sin_addr, sizeof(struct in_addr));
-                        }
-                        else {
-                                struct sockaddr_in6 *addr_in = (struct sockaddr_in6*) p_unicast->Address.lpSockaddr;
-                                memcpy(&(*mdns_ips).ipv6.addr, &addr_in->sin6_addr, sizeof(struct in6_addr));
-                        }
-                }
-                ++mdns_ips;
-                ++intfs;
+        *pp_mcast_addrs = mcast_addrs = malloc(sizeof(*mcast_addrs) * nb_intf);
+        if (mcast_addrs == NULL) {
+                free(mdns_ips);
+                free(intfs);
+                free(res);
+                return (MDNS_ERROR);
+        }
+        for ( const struct addrinfo* addr = addrs; addr != NULL; addr = addr->ai_next ) {
+            for (current = res; current != NULL; current = current->Next) {
+                    if (!mdns_is_interface_valuable(current, addr->ai_family))
+                            continue;
+                    if (addr->ai_family == AF_INET6) {
+                            *intfs = current->Ipv6IfIndex;
+                    }
+                    else {
+                            *intfs = current->IfIndex;
+                    }
+                    memcpy(mcast_addrs, addr->ai_addr, sa_len(addr->ai_addr));
+                    PIP_ADAPTER_UNICAST_ADDRESS p_unicast = current->FirstUnicastAddress;
+                    if (p_unicast == NULL) {
+                            free(mdns_ips);
+                            free(intfs);
+                            free(res);
+                            return (MDNS_ERROR);
+                    }
+                    // Take the first unicast address (highest priority)
+                    for (; p_unicast != NULL; p_unicast = p_unicast->Next) {
+                            if (p_unicast->Address.lpSockaddr->sa_family != addr->ai_family) {
+                                    continue;
+                            }
+
+                            if (addr->ai_family == AF_INET) {
+                                memcpy(mdns_ips, p_unicast->Address.lpSockaddr,
+                                    sizeof(struct sockaddr_in));
+                            }
+                            else {
+                                    memcpy(mdns_ips, p_unicast->Address.lpSockaddr,
+                                        sizeof(struct sockaddr_in6));
+                            }
+                    }
+                    ++mdns_ips;
+                    ++intfs;
+                    ++mcast_addrs;
+            }
         }
         *p_nb_intf = nb_intf;
         free(res);
@@ -289,21 +384,50 @@ mdns_resolve(struct mdns_ctx *ctx, const char *addr, unsigned short port)
 {
         char buf[6];
         struct addrinfo hints, *res = NULL;
-        multicast_if* ifaddrs = NULL;
-        struct mdns_ip *mdns_ips = NULL;
+        uint32_t* ifaddrs = NULL;
+        struct sockaddr_storage *mdns_ips = NULL;
+        struct sockaddr_storage *mcast_addrs = NULL;
         size_t i;
         int status;
 
         sprintf(buf, "%hu", port);
         memset(&hints, 0, sizeof(hints));
-        hints.ai_family = AF_UNSPEC;
         hints.ai_socktype = SOCK_DGRAM;
         hints.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV;
-        errno = getaddrinfo(addr, buf, &hints, &res);
-        if (errno != 0)
-                return (MDNS_LKPERR);
 
-        status = mdns_list_interfaces(&ifaddrs, &mdns_ips, &ctx->nb_conns, res->ai_family);
+        if (addr == NULL) {
+            hints.ai_family = AF_INET;
+            /* First, get the ipv4 multicast address info */
+            errno = getaddrinfo(MDNS_ADDR_IPV4, buf, &hints, &res);
+            if (errno != 0)
+                    return (MDNS_LKPERR);
+            /* Now get the ipv6 informations and link them with the ipv4 ones */
+            struct addrinfo *ipv6_res = NULL;
+            hints.ai_family = AF_INET6;
+            errno = getaddrinfo(MDNS_ADDR_IPV6, buf, &hints, &ipv6_res);
+            if (errno != 0) {
+                    freeaddrinfo(res);
+                    return (MDNS_LKPERR);
+            }
+            struct addrinfo* ipv4 = res;
+            for (; ipv4->ai_next != NULL; ipv4 = ipv4->ai_next)
+                ;
+            ipv4->ai_next = ipv6_res;
+        } else {
+            int family;
+            if (!strcmp(addr, MDNS_ADDR_IPV4))
+                family = AF_INET;
+            else if (!strcmp(addr, MDNS_ADDR_IPV6))
+                family = AF_INET6;
+            else
+                return (MDNS_LKPERR);
+            hints.ai_family = family;
+            errno = getaddrinfo(addr, buf, &hints, &res);
+            if (errno != 0)
+                    return (MDNS_LKPERR);
+        }
+
+        status = mdns_list_interfaces(&ifaddrs, &mdns_ips, &ctx->nb_conns, &mcast_addrs, res);
         if ( status < 0) {
                 freeaddrinfo(res);
                 return (status);
@@ -312,19 +436,21 @@ mdns_resolve(struct mdns_ctx *ctx, const char *addr, unsigned short port)
                 freeaddrinfo(res);
                 return (MDNS_NETERR);
         }
-        memcpy(&ctx->addr, res->ai_addr, res->ai_addrlen);
         ctx->conns = malloc(ctx->nb_conns * sizeof(*ctx->conns));
         if (ctx->conns == NULL) {
                 free(ifaddrs);
+                free(mdns_ips);
                 freeaddrinfo(res);
+                free(mcast_addrs);
                 return (MDNS_ERROR);
         }
         for (i = 0; i < ctx->nb_conns; ++i ) {
                 ctx->conns[i].sock = INVALID_SOCKET;
-                ctx->conns[i].if_addr = ifaddrs[i];
-                ctx->conns[i].mdns_ip = mdns_ips[i];
-                ctx->conns[i].mdns_ip.family = res->ai_family;
+                ctx->conns[i].intf_idx = ifaddrs[i];
+                ctx->conns[i].intf_addr = mdns_ips[i];
+                ctx->conns[i].mcast_addr = mcast_addrs[i];
         }
+        free(mcast_addrs);
         free(ifaddrs);
         free(mdns_ips);
         freeaddrinfo(res);
@@ -336,15 +462,13 @@ mdns_init(struct mdns_ctx **p_ctx, const char *addr, unsigned short port)
 {
         const uint32_t on_off = 1;
         const uint32_t ttl = 255;
-        const uint8_t loop = 1;
+        const uint32_t loop = 1;
         int res;
-#ifdef _WIN32
         union {
                 struct sockaddr_storage ss;
                 struct sockaddr_in      sin;
                 struct sockaddr_in6     sin6;
         } dumb;
-#endif /* _WIN32 */
         struct mdns_ctx *ctx;
 
         if (p_ctx == NULL)
@@ -366,14 +490,14 @@ mdns_init(struct mdns_ctx **p_ctx, const char *addr, unsigned short port)
                 return mdns_destroy(ctx), (res);
 
         for (size_t i = 0; i < ctx->nb_conns; ++i ) {
-                if ((ctx->conns[i].sock = socket(ctx->conns[i].mdns_ip.family, SOCK_DGRAM, IPPROTO_UDP)) == INVALID_SOCKET)
+                struct sockaddr_storage* ss_addr = (struct sockaddr_storage*)&ctx->conns[i].mcast_addr;
+                if ((ctx->conns[i].sock = socket(ctx->conns[i].intf_addr.ss_family, SOCK_DGRAM, IPPROTO_UDP)) == INVALID_SOCKET)
                         return mdns_destroy(ctx), (MDNS_NETERR);
                 if (setsockopt(ctx->conns[i].sock, SOL_SOCKET, SO_REUSEADDR, (const void *) &on_off, sizeof(on_off)) < 0)
                         return mdns_destroy(ctx), (MDNS_NETERR);
-    #ifdef _WIN32
             /* bind the receiver on any local address */
             memset(&dumb, 0, sizeof(dumb));
-            dumb.ss.ss_family = ss_family(&ctx->addr);
+            dumb.ss.ss_family = ss_family(&ctx->conns[i].intf_addr);
             if (dumb.ss.ss_family == AF_INET) {
                 dumb.sin.sin_port = htons(port);
                 dumb.sin.sin_addr.s_addr = INADDR_ANY;
@@ -381,37 +505,25 @@ mdns_init(struct mdns_ctx **p_ctx, const char *addr, unsigned short port)
                 dumb.sin6.sin6_port = htons(port);
                 dumb.sin6.sin6_addr = in6addr_any;
             }
+            if (bind(ctx->conns[i].sock, (const struct sockaddr *) &dumb,
+                     ss_len(&dumb.ss)) < 0)
+                    return mdns_destroy(ctx), (MDNS_NETERR);
 
-            if (bind(ctx->conns[i].sock, (const struct sockaddr *) &dumb, ss_len(&dumb.ss)) < 0)
+            if (os_mcast_join(ctx->conns[i].sock, ss_addr, ctx->conns[i].intf_idx,
+                              &ctx->conns[i].intf_addr) < 0)
                     return mdns_destroy(ctx), (MDNS_NETERR);
-#else /* _WIN32 */
-            if (bind(ctx->conns[i].sock, (const struct sockaddr *) &ctx->addr, ss_len(&ctx->addr)) < 0)
-                    return mdns_destroy(ctx), (MDNS_NETERR);
-#endif /* _WIN32 */
-
-            if (os_mcast_join(ctx->conns[i].sock, &ctx->addr, ctx->conns[i].if_addr) < 0)
-                    return mdns_destroy(ctx), (MDNS_NETERR);
-            if (setsockopt(ctx->conns[i].sock, ctx->conns[i].mdns_ip.family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
-                           ctx->conns[i].mdns_ip.family == AF_INET ? IP_MULTICAST_TTL : IPV6_MULTICAST_HOPS,
+            if (setsockopt(ctx->conns[i].sock, ss_level(&ctx->conns[i].intf_addr),
+                           ctx->conns[i].intf_addr.ss_family == AF_INET ? IP_MULTICAST_TTL : IPV6_MULTICAST_HOPS,
                            (const void *) &ttl, sizeof(ttl)) < 0) {
                     return mdns_destroy(ctx), (MDNS_NETERR);
             }
 
-            if (setsockopt(ctx->conns[i].sock, ctx->conns[i].mdns_ip.family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
-                           IP_MULTICAST_LOOP, (const void *) &loop, sizeof(loop)) < 0) {
+            if (setsockopt(ctx->conns[i].sock, ss_level(&ctx->conns[i].intf_addr),
+                           ctx->conns[i].intf_addr.ss_family == AF_INET ? IP_MULTICAST_LOOP : IPV6_MULTICAST_LOOP,
+                           (const void *) &loop, sizeof(loop)) < 0) {
                     return mdns_destroy(ctx), (MDNS_NETERR);
             }
-
-#if defined(HAVE_GETIFADDRS) || defined(_WIN32)
-            if (setsockopt(ctx->conns[i].sock,
-                           ctx->conns[i].mdns_ip.family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
-                           ctx->conns[i].mdns_ip.family == AF_INET ? IP_MULTICAST_IF : IPV6_MULTICAST_IF,
-                           (const void*)&ctx->conns[i].if_addr, sizeof(ctx->conns[i].if_addr))) {
-                    return mdns_destroy(ctx), (MDNS_NETERR);
-            }
-#endif
         }
-
         *p_ctx = ctx;
         return (0);
 }
@@ -433,7 +545,6 @@ mdns_destroy(struct mdns_ctx *ctx)
 
                         while ((svc = ctx->services)) {
                                 ctx->services = ctx->services->next;
-                                if (svc->name) free(svc->name);
                                 free(svc);
                         }
                 }
@@ -445,42 +556,59 @@ mdns_destroy(struct mdns_ctx *ctx)
 }
 
 static size_t
-mdns_write_hdr(uint8_t *ptr, const struct mdns_hdr *hdr)
+mdns_write_hdr(uint8_t *ptr, size_t* s, const struct mdns_hdr *hdr)
 {
         uint8_t *p = ptr;
 
-        p = write_u16(p, hdr->id);
-        p = write_u16(p, hdr->flags);
-        p = write_u16(p, hdr->num_qn);
-        p = write_u16(p, hdr->num_ans_rr);
-        p = write_u16(p, hdr->num_auth_rr);
-        p = write_u16(p, hdr->num_add_rr);
+        if (*s < 12)
+                return (-1);
+
+        p = write_u16(p, s, hdr->id);
+        p = write_u16(p, s, hdr->flags);
+        p = write_u16(p, s, hdr->num_qn);
+        p = write_u16(p, s, hdr->num_ans_rr);
+        p = write_u16(p, s, hdr->num_auth_rr);
+        p = write_u16(p, s, hdr->num_add_rr);
         return (p - ptr);
 }
 
+int
+mdns_write(const struct mdns_hdr *hdr, const struct rr_entry *entries,
+           uint8_t *buf, size_t bufSize, size_t* length)
+{
+    *length = 0;
+    if (!entries) return (MDNS_ERROR);
+    const struct rr_entry *entry = entries;
+    ssize_t l;
+
+    l = mdns_write_hdr(buf, &bufSize, hdr);
+    if (l < 0)
+            return (MDNS_ERROR);
+    *length += l;
+
+    for (entry = entries; entry; entry = entry->next) {
+            l = rr_write(buf + *length, &bufSize, entry, (hdr->flags & FLAG_QR) > 0);
+            if (l < 0) {
+                    return (MDNS_STDERR);
+            }
+            *length += l;
+    }
+    return (0);
+}
+
 int
 mdns_entries_send(const struct mdns_ctx *ctx, const struct mdns_hdr *hdr, const struct rr_entry *entries)
 {
         uint8_t buf[MDNS_PKT_MAXSZ] = {0};
-        const struct rr_entry *entry = entries;
-        ssize_t n = 0, l, r;
+        size_t l;
 
-        if (!entries) return (MDNS_ERROR);
-
-        l = mdns_write_hdr(buf, hdr);
-        n += l;
-
-        for (entry = entries; entry; entry = entry->next) {
-                l = rr_write(buf+n, entry, (hdr->flags & FLAG_QR) > 0);
-                if (l < 0) {
-                        return (MDNS_STDERR);
-                }
-                n += l;
-        }
+        if (mdns_write(hdr, entries, buf, sizeof(buf), &l) < 0)
+                return (MDNS_ERROR);
 
         for (size_t i = 0; i < ctx->nb_conns; ++i) {
-            r = sendto(ctx->conns[i].sock, (const char *) buf, n, 0,
-                    (const struct sockaddr *) &ctx->addr, ss_len(&ctx->addr));
+            ssize_t r = sendto(ctx->conns[i].sock, (const char *) buf, l, 0,
+                    (const struct sockaddr *) &ctx->conns[i].mcast_addr,
+                               ss_len((struct sockaddr_storage*)&ctx->conns[i].mcast_addr));
             if (r < 0)
                 return (MDNS_NETERR);
         }
@@ -488,7 +616,7 @@ mdns_entries_send(const struct mdns_ctx *ctx, const struct mdns_hdr *hdr, const
         return (0);
 }
 
-static void
+void
 mdns_free(struct rr_entry *entries)
 {
         struct rr_entry *entry;
@@ -501,58 +629,69 @@ mdns_free(struct rr_entry *entries)
 }
 
 static const uint8_t *
-mdns_read_header(const uint8_t *ptr, size_t n, struct mdns_hdr *hdr)
+mdns_read_header(const uint8_t *ptr, size_t *n, struct mdns_hdr *hdr)
 {
-        if (n <= sizeof(struct mdns_hdr)) {
+        if (*n <= sizeof(struct mdns_hdr)) {
                 errno = ENOSPC;
                 return NULL;
         }
-        ptr = read_u16(ptr, &n, &hdr->id);
-        ptr = read_u16(ptr, &n, &hdr->flags);
-        ptr = read_u16(ptr, &n, &hdr->num_qn);
-        ptr = read_u16(ptr, &n, &hdr->num_ans_rr);
-        ptr = read_u16(ptr, &n, &hdr->num_auth_rr);
-        ptr = read_u16(ptr, &n, &hdr->num_add_rr);
+        ptr = read_u16(ptr, n, &hdr->id);
+        ptr = read_u16(ptr, n, &hdr->flags);
+        ptr = read_u16(ptr, n, &hdr->num_qn);
+        ptr = read_u16(ptr, n, &hdr->num_ans_rr);
+        ptr = read_u16(ptr, n, &hdr->num_auth_rr);
+        ptr = read_u16(ptr, n, &hdr->num_add_rr);
         return ptr;
 }
 
+int
+mdns_parse(struct mdns_hdr *hdr, struct rr_entry **entries, const uint8_t *buf,
+           size_t length)
+{
+    size_t num_entry;
+    struct rr_entry *entry;
+
+    *entries = NULL;
+
+    const uint8_t *ptr = mdns_read_header(buf, &length, hdr);
+    if (ptr == NULL)
+            return (MDNS_ERROR);
+
+    num_entry = hdr->num_qn + hdr->num_ans_rr + hdr->num_add_rr;
+    for (size_t i = 0; i < num_entry; ++i) {
+            entry = calloc(1, sizeof(struct rr_entry));
+            if (!entry)
+                    goto err;
+            ptr = rr_read(ptr, &length, buf, entry, i >= hdr->num_qn);
+            if (!ptr) {
+                    mdns_free(entry);
+                    errno = ENOSPC;
+                    goto err;
+            }
+            entry->next = *entries;
+            *entries = entry;
+    }
+    if (*entries == NULL) {
+            return (MDNS_ERROR);
+    }
+    return (0);
+err:
+    mdns_free(*entries);
+    *entries = NULL;
+    return (MDNS_ERROR);
+}
+
 static int
 mdns_recv(const struct mdns_conn* conn, struct mdns_hdr *hdr, struct rr_entry **entries)
 {
         uint8_t buf[MDNS_PKT_MAXSZ];
-        size_t num_entry, n;
         ssize_t length;
-        struct rr_entry *entry;
 
         *entries = NULL;
         if ((length = recv(conn->sock, (char *) buf, sizeof(buf), 0)) < 0)
                 return (MDNS_NETERR);
 
-        const uint8_t *ptr = mdns_read_header(buf, length, hdr);
-        n = length;
-
-        num_entry = hdr->num_qn + hdr->num_ans_rr + hdr->num_add_rr;
-        for (size_t i = 0; i < num_entry; ++i) {
-                entry = calloc(1, sizeof(struct rr_entry));
-                if (!entry)
-                        goto err;
-                ptr = rr_read(ptr, &n, buf, entry, i >= hdr->num_qn);
-                if (!ptr) {
-                        free(entry);
-                        errno = ENOSPC;
-                        goto err;
-                }
-                entry->next = *entries;
-                *entries = entry;
-        }
-        if (*entries == NULL) {
-                return (MDNS_ERROR);
-        }
-        return (0);
-err:
-        mdns_free(*entries);
-        *entries = NULL;
-        return (MDNS_ERROR);
+        return mdns_parse(hdr, entries, buf, (size_t)length);
 }
 
 void
@@ -685,7 +824,7 @@ mdns_listen(const struct mdns_ctx *ctx, const char *const names[],
 }
 
 int
-mdns_announce(struct mdns_ctx *ctx, const char *service, enum rr_type type,
+mdns_announce(struct mdns_ctx *ctx, enum rr_type type,
         mdns_announce_callback callback, void *p_cookie)
 {
         if (!callback)
@@ -695,7 +834,6 @@ mdns_announce(struct mdns_ctx *ctx, const char *service, enum rr_type type,
         if (!svc)
                 return (MDNS_ERROR);
 
-        svc->name = strdup(service);
         svc->type = type;
         svc->announce_callback = callback;
         svc->p_cookie = p_cookie;
@@ -705,6 +843,19 @@ mdns_announce(struct mdns_ctx *ctx, const char *service, enum rr_type type,
         return (0);
 }
 
+void
+mdns_request_initial_announce(struct mdns_ctx *ctx, const char* service)
+{
+    /* Send the initial announce (RFC 6762 §8.3) */
+    for (struct mdns_svc* svc = ctx->services; svc; svc = svc->next) {
+        for ( size_t i = 0; i < ctx->nb_conns; ++i ) {
+            svc->announce_callback(svc->p_cookie,
+                                   (struct sockaddr*)&ctx->conns[i].intf_addr,
+                                   service, MDNS_ANNOUNCE_INITIAL);
+        }
+    }
+}
+
 int
 mdns_serve(struct mdns_ctx *ctx, mdns_stop_func stop, void *p_cookie)
 {
@@ -718,6 +869,15 @@ mdns_serve(struct mdns_ctx *ctx, mdns_stop_func stop, void *p_cookie)
                         return (MDNS_NETERR);
         }
 
+        /* Send the initial announce (RFC 6762 §8.3) */
+        for (svc = ctx->services; svc; svc = svc->next) {
+            for ( size_t i = 0; i < ctx->nb_conns; ++i ) {
+                svc->announce_callback(svc->p_cookie,
+                                       (struct sockaddr*)&ctx->conns[i].intf_addr,
+                                       NULL, MDNS_ANNOUNCE_INITIAL);
+            }
+        }
+
         for (; stop(p_cookie) == false;) {
                 struct pollfd *pfd = alloca( sizeof(*pfd) * ctx->nb_conns );
 
@@ -732,21 +892,33 @@ mdns_serve(struct mdns_ctx *ctx, mdns_stop_func stop, void *p_cookie)
                         if ((pfd[i].revents & POLLIN) == 0)
                                 continue;
                         r = mdns_recv(&ctx->conns[i], &qhdr, &question);
-                        if (r == MDNS_NETERR)
+                        if (r < 0)
                                 continue;
                         if (qhdr.num_qn == 0)
                                 goto again;
 
                         for (svc = ctx->services; svc; svc = svc->next) {
-                                if (!strrcmp(question->name, svc->name) && question->type == svc->type) {
-                                        svc->announce_callback(svc->p_cookie, r, &ctx->conns[i].mdns_ip, question);
+                                if (question->type == svc->type) {
+                                        svc->announce_callback(svc->p_cookie,
+                                               (struct sockaddr*)&ctx->conns[i].intf_addr,
+                                               question->name, MDNS_ANNOUNCE_RESPONSE);
                                         goto again;
                                 }
                         }
+                        mdns_free(question);
+                        question = NULL;
                 }
 again:
                 mdns_free(question);
                 question = NULL;
         }
+        /* Send the goodbye packets (RFC 6762 §10.1) */
+        for (svc = ctx->services; svc; svc = svc->next) {
+            for ( size_t i = 0; i < ctx->nb_conns; ++i ) {
+                svc->announce_callback(svc->p_cookie,
+                                       (struct sockaddr*)&ctx->conns[i].intf_addr, NULL,
+                                       MDNS_ANNOUNCE_GOODBYE);
+            }
+        }
         return (0);
 }
